-- Language Resource (version 1.0)

--    If you have questions you can contact me at:  m.schut@twi.tudelft.nl or
--    at m2s-soft@kagi.com
--
-- )))-----------------------------------------------------------------------(((
--
--    DISCLAIMER:  This  library  is offered free of charge,  as-is,  with no
--    warranties whatsoever, including implied warranties of merchantability.
--    I'm not responsible for any use or misuse. Use this at your own risk.
--
-- )))-----------------------------------------------------------------------(((
--
--    Copyright 1999 by Martin Schut / M2S-SOFT.


----------------------------
-- The hashfunction include is the same as in the 'Hash table Demo'
-- included with the Euphoria distribution package.
----------------------------

--with trace

--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--
-- Private (local) declarations, procedures and functions.
--

constant TRUE = 1                                                                       -- Widely used boolean constant
constant KEY = 1, RECORD = 2                                                    -- Fields for one hash table entry
constant HASH_BUCKETS = 1009                                            -- Hashtable size

constant ERROR_READING_KEYS = 1                                         -- Error reading key-file
constant ERROR_READING_VALUES = 2                                               -- Error reading language-file
constant ERROR_KEY_EXIST = 3                                                    -- Key already exist
constant ERROR_NO_LANGUAGE = 4                                          -- No language selected
constant ERROR_RES_NOT_FOUND = 5                                                -- Resource not found
constant ERROR_KEY_NOT_FOUND = 6                                                -- Key not found

atom record record = 1                                                          -- Current number of records
integer traceError traceError = 0                                               -- Trace
sequence keys keys = repeat({}, HASH_BUCKETS)                   -- The resource keys
sequence values values = {}                                                     -- The current tranlations
sequence language language = ""                                         -- The current language
sequence resourceDir resourceDir = ""                                   -- The resource-directory
sequence resourceFiles resourceFiles = {}                               -- The resource-files

function nlErase(object string)
	-- Erase the newline character
	if atom(string) then return string end if

	if string[length(string)] = 10 then
		return string[1..length(string) - 1]
	else
		return string
	end if
end function

function trim(object string)
	-- Erase the spaces from the beginning and end of a sequence
	integer first, last

	if atom(string) then return string end if

	first = 0
	last = 1
	for i=1 to length(string) do
		if string[i] != ' ' then
			first = i
			exit
		end if
	end for

	for i=length(string) to 1 by -1 do
		if string[i] != ' ' then
			last = i
			exit
		end if
	end for

	return string[first..last]
end function

procedure resError(integer error, sequence file, sequence key)
	sequence message

	if traceError then        
		message = "*** ERROR ***\n"
		if error = ERROR_READING_KEYS then
			message &= "Couldn't read the specified key-file. (" & resourceDir & file & ".key)\n"
		elsif error = ERROR_READING_VALUES then
			message &= "Couldn't read the specified language-file. (" & resourceDir & file & "." & language & ")\n"
		elsif error = ERROR_KEY_EXIST then
			message &= "Couldn't insert key (" & key & ") from key-file (" & resourceDir & file & ".key). This key already exist\n"
		elsif error = ERROR_NO_LANGUAGE then
			message &= "No language specified."
		elsif error = ERROR_RES_NOT_FOUND then
			message &= "Resource " & file & " not specified. Can't remove resource."
		elsif error = ERROR_KEY_NOT_FOUND then
			message &= "Key " & key & " not found."
		end if

		message &= "\n"

		puts(2, message)
		trace(traceError)
	end if
end procedure

function hash(sequence key)
	-- Calculate the hash-value.
	-- English words will give the best result.
	integer len
	atom val
    
	len = length(key)
	val = key[len]
	if len > 4 then len = 4 end if
	for i = 1 to len - 1 do
		val = val * 64 + key[i]                                         -- Shift 6 bits and add
	end for
	return remainder(val, HASH_BUCKETS) + 1
end function

function hashFind(sequence key)
	-- Return the recordnumber belonging to the key
	sequence bucket

	bucket = keys[hash(key)]                                                        -- Which bucket to search?
	for i = 1 to length(bucket) do                                  -- Search this bucket for the key
		if equal(key, bucket[i][KEY]) then                              -- Found it
			return bucket[i][RECORD]                                        -- Return the recordnumber
		end if
	end for
	return 0                                                                                -- Key not found
end function

procedure hashAdd(sequence key, integer record)
	-- Add a key to the hashtable.
	integer hash_val
	sequence bucket
    
	hash_val = hash(key)                                                    -- Which bucket to append to?
	bucket = keys[hash_val]
	bucket = append(bucket, {key, record})                          -- Append key to bucket
	keys[hash_val] = bucket                                                 -- 'Rewrite' bucket
end procedure

procedure hashRemove(sequence key)
	-- Remove a key from the hashtable.
	integer hash_val
	sequence tempBucket, bucket

	hash_val = hash(key)                                                    -- Which bucket to remove from?
	bucket = keys[hash_val]

	for i = 1 to length(bucket) do                                  -- Search this bucket for the key
		if equal(key, bucket[i][KEY]) then                              -- Found it
			tempBucket = repeat({}, length(bucket) - 1)
			tempBucket[1..i - 1] = bucket[1..i - 1]         -- Remove key
			tempBucket[i..length(bucket) - 1] = bucket[i + 1..length(bucket)]
			bucket = tempBucket
		end if
	end for
	keys[hash_val] = bucket                                                 -- 'Rewrite' bucket
end procedure

procedure loadLanguageFile(sequence filename)
	integer file
	object value

	file = open(resourceDir & filename & "." & language, "r")       -- Open the language-file
	if file = -1 then                                                               -- Error opening keys-file
		resError(ERROR_READING_VALUES, filename, "")
	end if
	while TRUE do
		value = nlErase(trim(gets(file)))
		if atom(value) then                                                     -- End language-file
			exit
		elsif not equal(value, "") then
			values = append(values, value)                  -- To current translation-list
		end if
	end while
	close(file)                                                                     -- Close the keys-file
end procedure


--
-- Private (local) declarations, procedures and functions.
--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--


--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--
-- Public (global) declarations, procedures and functions. Interface to the user.
--

global function enumResourceFiles()
	-- Return the list if resourceFile
	return resourceFiles
end function

global function isResIncluded(sequence filename)
	-- Return whether the specified file is included in
	-- the resources.
	return find(filename, resourceFiles) != 0
end function

global function getResourceDir()
	-- Get the current resource-directory
	return resourceDir
end function

global function getLanguage()
	-- Get the current language
	return language
end function

global procedure addResource(sequence filename)
	-- Add a resource to the translation-list
	integer file
	object key

	if equal(language, "") then                                             -- No language selected
		resError(ERROR_NO_LANGUAGE, "", "")
	end if
	if not isResIncluded(filename) then                             -- Resource not included
		file = open(resourceDir & filename & ".key", "r")       -- Open the keys-file
		if file = -1 then                                                       -- Error opening keys-file
			resError(ERROR_READING_KEYS, filename, "")
		end if
		while TRUE do
			key = nlErase(trim(gets(file)))
			if atom(key) then                                               -- End of keys-file
				exit
			elsif not equal(key, "") and not hashFind(key) then     -- Add {key, recordnumber} if the
				hashAdd(key, record)                            -- key does not exist
				record += 1
			else                                                                    -- Otherwise give a error
				resError(ERROR_KEY_EXIST, filename, key)
			end if
		end while
		close(file)                                                             -- Close the keys-file

		loadLanguageFile(filename)                                      -- Load the translations
		resourceFiles = append(resourceFiles, filename) -- Add to the translation files
	end if
end procedure

global procedure removeResource(sequence filename)
	-- Remove a resource from the translation-list
	integer file, rec, low, high
	object key
	sequence tempValues

	low = -1
	high = -1
	if equal(language, "") then                                             -- No language selected
		resError(ERROR_NO_LANGUAGE, "", "")
	elsif not isResIncluded(filename) then                          -- Resource not included
		resError(ERROR_RES_NOT_FOUND, filename, "")
	else
		file = find(filename, resourceFiles)
		tempValues = repeat({}, length(resourceFiles) - 1)
		tempValues[1..file - 1] = resourceFiles[1..file - 1]
		tempValues[file..length(resourceFiles) - 1] = resourceFiles[file + 1..length(resourceFiles)]
		resourceFiles = tempValues
	end if

	file = open(resourceDir & filename & ".key", "r")               -- Open the keys-file
	if file = -1 then                                                               -- Error opening keys-file
		resError(ERROR_READING_KEYS, filename, "")
	end if
	while TRUE do
		key = nlErase(trim(gets(file)))
		if atom(key) then exit end if                                   -- End of keys-file
		rec = hashFind(key)
		if not equal(key, "") and (rec > 0) then
			hashRemove(key)
			if low = -1 or rec < low then                           -- This record is the lowest in the sequence
				low = rec
			end if
			if high = -1 or rec > high then                 -- This record is the highest in the sequence
				high = rec
			end if
		else                                                                            -- Otherwise give a error
			resError(ERROR_KEY_NOT_FOUND, "", key)
		end if
	end while
	close(file)                                                                     -- Close the keys-file

	tempValues = repeat({}, low + (length(values) - high - 1))
	tempValues[1..low - 1] = values[1..low - 1]                     -- Erase translations
	tempValues[low..low + (length(values) - high - 1)] = values[high + 1..length(values)]
	values = tempValues
	record = length(values) + 1
end procedure

global procedure removeAllResources()
	-- Remove all resources

	keys = repeat({}, HASH_BUCKETS)                                 -- Clear hashtable
	values = {}                                                                     -- Clear translations
	record = 1
end procedure

global procedure setLanguage(sequence lang)
	-- Set the language
	if not equal(lang, "") and not equal(lang, language) then
		values = {}                                                             -- Clear the translation list
		language = lang                                                 -- Set the new language

		for i = 1 to length(resourceFiles) do                   -- Reload all translation
			loadLanguageFile(resourceFiles[i])
		end for
	end if
end procedure

global procedure setResourceDir(sequence dir)
	-- Set the resource-directory
	sequence tempResourceFiles

	if not equal(dir, resourceDir) then
		if not equal(dir, "") then                                      -- Add '\' if needed
			dir &= '\\'
		end if
		resourceDir = dir                                                       -- Set new resource-drectory
		keys = repeat({}, HASH_BUCKETS)                         -- Clear hashtable
		values = {}                                                             -- Clear the translation list
		tempResourceFiles = resourceFiles
		for i = 1 to length(resourceFiles) do                   -- Reload all resources
			addResource(tempResourceFiles[i])
		end for
	end if
end procedure

global function translate(sequence key)
	-- Translate a key into a word
	integer rec

	rec = hashFind(key)
	if not equal(key, "") and (rec > 0) then
		return values[rec]                                                      -- Return the translation
	else
		resError(ERROR_KEY_NOT_FOUND, "", key)                  -- No translation found
		return ""                                                                       -- Return key
	end if
end function

global procedure res_trace(integer t)
	traceError = t
end procedure

--
-- Public (global) declarations, procedures and functions. Interface to the user.
--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--o--
